#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2024, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
It contains a comprehensive list of all modules and classes within schedula.

Docstrings should provide sufficient understanding for any individual function.

Modules:

.. currentmodule:: schedula

.. autosummary::
    :nosignatures:
    :toctree: toctree/schedula

    ~dispatcher
    ~utils
    ~ext
    ~cli
"""
import os
import sys
from ._version import *

_all = {
    'Dispatcher': '.dispatcher',
    'BlueDispatcher': '.utils.blue',
    'Blueprint': '.utils.blue',
    'PoolExecutor': '.utils.asy.executors',
    'ProcessExecutor': '.utils.asy.executors',
    'ThreadExecutor': '.utils.asy.executors',
    'ProcessPoolExecutor': '.utils.asy.executors',
    'register_executor': '.utils.asy',
    'shutdown_executor': '.utils.asy',
    'shutdown_executors': '.utils.asy',
    'await_result': '.utils.asy',
    'EMPTY': '.utils.cst',
    'START': '.utils.cst',
    'NONE': '.utils.cst',
    'SINK': '.utils.cst',
    'SELF': '.utils.cst',
    'END': '.utils.cst',
    'PLOT': '.utils.cst',
    'stlp': '.utils.dsp',
    'combine_dicts': '.utils.dsp',
    'bypass': '.utils.dsp',
    'summation': '.utils.dsp',
    'map_dict': '.utils.dsp',
    'map_list': '.utils.dsp',
    'selector': '.utils.dsp',
    'replicate_value': '.utils.dsp',
    'partial': '.utils.dsp',
    'add_args': '.utils.dsp',
    'stack_nested_keys': '.utils.dsp',
    'get_nested_dicts': '.utils.dsp',
    'inf': '.utils.dsp',
    'are_in_nested_dicts': '.utils.dsp',
    'SubDispatchFunction': '.utils.dsp',
    'combine_nested_dicts': '.utils.dsp',
    'SubDispatch': '.utils.dsp',
    'run_model': '.utils.dsp',
    'MapDispatch': '.utils.dsp',
    'parent_func': '.utils.dsp',
    'SubDispatchPipe': '.utils.dsp',
    'DispatchPipe': '.utils.dsp',
    'kk_dict': '.utils.dsp',
    'add_function': '.utils.dsp',
    'DispatcherError': '.utils.exc',
    'DispatcherAbort': '.utils.exc',
    'ExecutorShutdown': '.utils.exc',
    'WebResponse': '.utils.exc',
    'SkipNode': '.utils.exc',
    'counter': '.utils.gen',
    'Token': '.utils.gen',
    'DiGraph': '.utils.graph',
    'save_dispatcher': '.utils.io',
    'load_dispatcher': '.utils.io',
    'save_default_values': '.utils.io',
    'load_default_values': '.utils.io',
    'save_map': '.utils.io',
    'load_map': '.utils.io'
}

__all__ = tuple(_all)


def __dir__():
    return __all__ + (
        '__doc__', '__author__', '__updated__', '__title__', '__version__',
        '__license__', '__copyright__'
    )


def __getattr__(name):
    if name in _all:
        import importlib
        obj = getattr(importlib.import_module(_all[name], __name__), name)
        globals()[name] = obj
        return obj
    raise AttributeError("module %s has no attribute %s" % (__name__, name))


if sys.version_info[:2] < (3, 7) or os.environ.get('IMPORT_ALL') == 'True':
    from .dispatcher import Dispatcher
    from .utils.asy import (
        await_result, register_executor, shutdown_executor, shutdown_executors
    )
    from .utils.asy.executors import (
        PoolExecutor, ProcessExecutor, ThreadExecutor, ProcessPoolExecutor
    )
    from .utils.blue import BlueDispatcher, Blueprint
    from .utils.cst import EMPTY, END, NONE, PLOT, SELF, SINK, START
    from .utils.dsp import (
        DispatchPipe, SubDispatch, SubDispatchFunction, SubDispatchPipe,
        MapDispatch, add_args, partial, run_model,
        add_function, are_in_nested_dicts, bypass, combine_dicts,
        combine_nested_dicts, get_nested_dicts, inf, kk_dict, map_dict,
        map_list,
        parent_func, replicate_value, selector, stack_nested_keys, stlp,
        summation
    )
    from .utils.exc import (
        DispatcherAbort, DispatcherError, ExecutorShutdown, SkipNode,
        WebResponse
    )
    from .utils.gen import Token, counter
    from .utils.graph import DiGraph

    try:
        from .utils.io import (
            load_default_values, load_dispatcher, load_map, save_default_values,
            save_dispatcher, save_map
        )
    except ImportError:  # MicroPython.
        pass



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2024, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides the `gunicorn` BaseApplication to run the server in production.
"""
import sys
import runpy
import functools
import os.path as osp
import multiprocessing
from gunicorn.app.base import BaseApplication


@functools.lru_cache()
def _get_module(module, path):
    return runpy.run_module(module)


def get_module(module=None, extra_sys_paths=()):
    gbl = {}
    if module:
        path = sys.path.copy() if extra_sys_paths else sys.path
        try:
            for d in extra_sys_paths[::-1]:
                sys.path.insert(0, osp.abspath(d))
            cpath = tuple(sys.path)
            for v in ([module] if isinstance(module, str) else module):
                gbl.update(_get_module(v, cpath))
        finally:
            sys.path = path
    return gbl


class Application(BaseApplication):
    def __init__(self, app, workers=None, timeout=0, threads=10,
                 accesslog='-', **options):
        self.options = {
            'threads': threads,
            'timeout': timeout,
            'workers': workers or (multiprocessing.cpu_count() * 2) + 1,
            'accesslog': accesslog,
            **options
        }
        self.application = app
        super().__init__()

    def load_config(self):
        invalid = {key for key in self.options if key not in self.cfg.settings}
        if invalid:
            raise ValueError(f'Invalid gunicorn options `{", ".join(invalid)}`.')
        config = {
            key: value for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2024, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It defines the `ExecutorFactory` class.
"""
from ..cst import EMPTY
from ..imp import finalize, Lock
from ..dsp import get_nested_dicts


class ExecutorFactory(dict):
    def __init__(self, *args, **kwargs):
        super(ExecutorFactory, self).__init__(*args, **kwargs)
        self._executors = {}
        self._lock = Lock()
        finalize(self, self.shutdown_executor, wait=False)

    def __getstate__(self):
        it = self.__dict__.items()
        return {k: v for k, v in it if k != '_lock' and k != '_executors'}

    def __setstate__(self, state):
        self.__dict__ = state
        self._executors = {}
        self._lock = Lock()
        finalize(self, self.shutdown_executor, wait=False)

    @staticmethod
    def executor_id(name, sol):
        if name is True:
            name = sol.dsp.executor
        return name, id(sol)

    def get_executor(self, exe_id):
        name, sol_id = exe_id
        if name is not False:
            d = get_nested_dicts(self._executors, name)
            if get_nested_dicts(d, 'active', sol_id, default=lambda: True):
                default = self.get(name, lambda: None)
                with self._lock:
                    return get_nested_dicts(d, 'executor', default=default)
            else:
                from .executors import PoolExecutor
                # noinspection PyTypeChecker
                return PoolExecutor(None)

    def set_executor(self, name, value):
        get_nested_dicts(self._executors, name)['executor'] = value

    def _filter_executors(self, name=EMPTY, sol_id=EMPTY):
        bn, bs = name is EMPTY, sol_id is EMPTY
        for n, d in self._executors.items():
            if (bn or n == name) and (bs or sol_id in d['active']):
                yield n, d

    def set_active(self, sol_id, value=True):
        for k, d in self._filter_executors(sol_id=sol_id):
            d['active'][sol_id] = value

    def pop_active(self, sol_id):
        for k, d in self._filter_executors(sol_id=sol_id):
            d['active'].pop(sol_id, None)

    def shutdown_executor(self, name=EMPTY, sol_id=EMPTY, wait=True):
        data = dict(self._filter_executors(name=name, sol_id=sol_id))
        if wait:
            from concurrent.futures import wait as _wait_fut
            futures = set()
            for d in data.values():
                if d.get('executor'):
                    futures.update(d['executor'].get_futures(sol_id))
            # noinspection PyCallingNonCallable,PyUnboundLocalVariable
            _wait_fut(futures)

        with self._lock:
            tasks, force = {}, sol_id is EMPTY
            for k, d in self._filter_executors(name=name, sol_id=sol_id):
                active = d.get('active', {})
                active[sol_id] = False
                if force or not any(active.values()):
                    active.update(dict.fromkeys(active, False))
                    tasks[k] = d.pop('executor', None)

        return {k: e.shutdown(0) for k, e in tasks.items() if e}



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2024, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides the default Flask App config file.
"""
import os
import secrets
import inspect
import os.path as osp

try:
    import flask_security
except ImportError:
    flask_security = None


class Config:
    DEBUG = True
    WTF_CSRF_CHECK_DEFAULT = False

    # reCAPTCHA configuration
    RECAPTCHA_PUBLIC_KEY = os.environ.get(
        'RECAPTCHA_PUBLIC_KEY', '6LcsgJglAAAAAMm7ilxkhBRevaCAuxlpefYZmxHU'
    )
    RECAPTCHA_PRIVATE_KEY = os.environ.get(
        'RECAPTCHA_PRIVATE_KEY', '6LcsgJglAAAAAAbR3aHm2qJS_c3XsGqmC9O816eH'
    )

    # Generate a nice key using secrets.token_urlsafe()
    SECRET_KEY = os.environ.get("SECRET_KEY", secrets.token_urlsafe())
    # Bcrypt is set as default SECURITY_PASSWORD_HASH, which requires a salt
    # Generate a good salt using: secrets.SystemRandom().getrandbits(128)
    SECURITY_PASSWORD_SALT = os.environ.get(
        "SECURITY_PASSWORD_SALT", f'{secrets.SystemRandom().getrandbits(128)}'
    )
    SECURITY_BLUEPRINT_NAME = os.environ.get(
        "SECURITY_BLUEPRINT_NAME", 'security'
    )
    SECURITY_URL_PREFIX = os.environ.get("SECURITY_URL_PREFIX", '/user')
    SECURITY_CONFIRMABLE = os.environ.get(
        "SECURITY_CONFIRMABLE", 'true'
    ).lower() == 'true'
    SECURITY_AUTO_LOGIN_AFTER_CONFIRM = os.environ.get(
        "SECURITY_AUTO_LOGIN_AFTER_CONFIRM", 'false'
    ).lower() == 'true'
    SECURITY_POST_CONFIRM_VIEW = os.environ.get(
        "SECURITY_POST_CONFIRM_VIEW", '/#login'
    )
    SECURITY_CONFIRM_ERROR_VIEW = os.environ.get(
        "SECURITY_CONFIRM_ERROR_VIEW", '/#login'
    )
    SECURITY_REGISTERABLE = os.environ.get(
        "SECURITY_REGISTERABLE", 'true'
    ).lower() == 'true'
    SECURITY_SEND_REGISTER_EMAIL = os.environ.get(
        "SECURITY_SEND_REGISTER_EMAIL", 'true'
    ).lower() == 'true'
    SECURITY_RECOVERABLE = os.environ.get(
        "SECURITY_RECOVERABLE", 'true'
    ).lower() == 'true'
    SECURITY_RESET_VIEW = os.environ.get("SECURITY_RESET_VIEW", '/#reset')
    SECURITY_RESET_ERROR_VIEW = os.environ.get(
        "SECURITY_RESET_ERROR_VIEW", '/#login'
    )
    SECURITY_REDIRECT_BEHAVIOR = os.environ.get(
        "SECURITY_REDIRECT_BEHAVIOR", 'spa'
    )

    SECURITY_TRACKABLE = os.environ.get(
        "SECURITY_TRACKABLE", 'true'
    ).lower() == 'true'

    MAIL_DEFAULT_SENDER = os.environ.get('MAIL_DEFAULT_SENDER')
    MAIL_SERVER = os.environ.get('MAIL_HOST')
    MAIL_PORT = int(os.environ.get('MAIL_PORT', 465))
    MAIL_USE_SSL = True
    MAIL_USERNAME = os.environ.get('MAIL_USER')
    MAIL_PASSWORD = os.environ.get('MAIL_PASSWORD')

    SECURITY_I18N_DIRNAME = [
        "translations",
        os.environ.get('SECURITY_I18N_DIRNAME', 'translations'),
        osp.join(osp.dirname(osp.dirname(__file__)), 'translations')
    ]
    if flask_security is not None:
        SECURITY_I18N_DIRNAME.append(osp.join(
            osp.dirname(inspect.getfile(flask_security)), 'translations'
        ))

    SCHEDULA_I18N_DIRNAME = [
        "translations",
        os.environ.get('SCHEDULA_I18N_DIRNAME', 'translations'),
        osp.join(osp.dirname(__file__), 'translations'),
    ]

    BABEL_DEFAULT_LOCALE = 'en_US'
    BABEL_LANGUAGES = {
        'af_ZA': {"icon": "🇿🇦", "label": "Afrikaans"},
        'ca_ES': {"icon": "🇪🇸", "label": "Català"},
        'da_DK': {"icon": "🇩🇰", "label": "Dansk"},
        'de_DE': {"icon": "🇩🇪", "label": "Deutsch"},
        'en_US': {"icon": "🇺🇸", "label": "English"},
        'es_ES': {"icon": "🇪🇸", "label": "Español"},
        'eu_ES': {"icon": "🇪🇸", "label": "Euskara"},
        'fr_FR': {"icon": "🇫🇷", "label": "Français"},
        'hu_HU': {"icon": "🇭🇺", "label": "Magyar"},
        'hy_AM': {"icon": "🇦🇲", "label": "Հայերեն"},
        'is_IS': {"icon": "🇮🇸", "label": "Íslenska"},
        'it_IT': {"icon": "🇮🇹", "label": "Italiano"},
        'ja_JP': {"icon": "🇯🇵", "label": "日本語"},
        'nl_NL': {"icon": "🇳🇱", "label": "Nederlands"},
        'pl_PL': {"icon": "🇵🇱", "label": "Polski"},
        'pt_BR': {"icon": "🇧🇷", "label": "Português (Brasil)"},
        'pt_PT': {"icon": "🇵🇹", "label": "Português (Portugal)"},
        'ru_RU': {"icon": "🇷🇺", "label": "Русский"},
        'tr_TR': {"icon": "🇹🇷", "label": "Türkçe"},
        'zh_Hans_CN': {"icon": "🇨🇳", "label": "中文（简体）"}
    }

    # have session and remember cookie be samesite (flask/flask_login)
    REMEMBER_COOKIE_SAMESITE = "strict"
    SESSION_COOKIE_SAMESITE = "strict"

    # Use an in-memory db
    SQLALCHEMY_DATABASE_URI = 'sqlite://'
    # As of Flask-SQLAlchemy 2.4.0 it is easy to pass in options directly to the
    # underlying engine. This option makes sure that DB connections from the
    # pool are still valid. Important for entire application since
    # many DBaaS options automatically close idle connections.
    SQLALCHEMY_ENGINE_OPTIONS = {"pool_pre_ping": True}
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    STRIPE_SECRET_KEY = os.environ.get("STRIPE_SECRET_KEY")
    STRIPE_PUBLISHABLE_KEY = os.environ.get("STRIPE_PUBLISHABLE_KEY")
    STRIPE_WEBHOOK_SECRET_KEY = os.environ.get("STRIPE_WEBHOOK_SECRET_KEY")



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2024, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to dump and load secrets from flask session when dealing
with JSON.
"""

import json
import hashlib
from flask import session


def dumps_secret(o):
    dhash = hashlib.sha256()
    dhash.update(json.dumps(o, sort_keys=True).encode())
    key = dhash.hexdigest()
    if not '$secrets' in session:
        session['$secrets'] = {}
    if key not in session['$secrets']:
        session['$secrets'][key] = o
    return key


def loads_secret(key):
    return session['$secrets'][key]


def secrets(obj, dumps=True):
    if isinstance(obj, list):
        return [secrets(v, dumps) for v in obj]
    elif isinstance(obj, dict):
        res = {
            k: secrets(v, dumps)
            for k, v in obj.items() if '$secret' != k
        }
        if '$secret' in obj:
            if dumps:
                res['$secret'] = dumps_secret(obj['$secret'])
            else:
                obj = loads_secret(obj['$secret'])
                if res:
                    res.update(obj)
                else:
                    res = obj
        return res
    return obj


def dumps(obj, **kwargs):
    return json.dumps(secrets(obj), **kwargs)


def loads(s, **kwargs):
    return secrets(json.loads(s, **kwargs), dumps=False)



# coding=utf-8
# -*- coding: UTF-8 -*-
#
# Copyright 2015-2024, Vincenzo Arcidiacono;
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

"""
It provides functions to build the base form flask app.
"""
import logging
import datetime
import collections
import os.path as osp
import schedula as sh
from .mail import Mail
from . import json_secrets
from .config import Config
from flask_sqlalchemy import SQLAlchemy
from flask_babel import Babel, lazy_gettext
from werkzeug.datastructures import MultiDict
from flask_wtf.recaptcha import RecaptchaField
from flask_principal import Permission, RoleNeed
from flask_security.models import fsqla_v3 as fsqla
from flask import after_this_request, request, jsonify, current_app, flash
from sqlalchemy import Column, String, Integer, DateTime, ForeignKey, JSON
from flask_security.utils import (
    base_render_json, suppress_form_csrf, view_commit
)
from flask_security.forms import (
    ConfirmRegisterForm, Required, StringField, Form, EmailField
)
from flask_security import (
    Security, SQLAlchemyUserDatastore, current_user as cu, auth_required
)

log = logging.getLogger(__name__)


def default_get_form_context():
    return {
        'userInfo': getattr(cu, "get_security_payload", lambda: {})(),
        'reCAPTCHA': current_app.config.get('RECAPTCHA_PUBLIC_KEY'),
        'stripeKey': current_app.config.get('STRIPE_PUBLISHABLE_KEY')
    }


def basic_app(sitemap, app):
    app.config.from_object(Config)
    if getattr(sitemap, 'basic_app_config'):
        app.config.from_object(sitemap.basic_app_config)

    # Create database connection object
    db = SQLAlchemy(app)

    def default_name(context):
        return f'Item {context.get_current_parameters()["id"]}'

    def is_admin():
        return Permission(RoleNeed('admin')).can()

    class Item(db.Model):
        __tablename__ = 'item'
        id = Column(Integer, primary_key=True)
        name = Column(String(255), default=default_name)
        category = Column(String(255))
        data = Column('data', JSON)
        user_id = Column(Integer, ForeignKey('user.id'))
        created_at = Column(DateTime(), default=datetime.datetime.utcnow)
        updated_at = Column(DateTime(), onupdate=datetime.datetime.utcnow)

        def payload(self, data=False):
            res = {
                'id': self.id,
                'name': self.name,
                'created_at': self.created_at,
                'updated_at': self.updated_at
            }
            for k in ('created_at', 'updated_at'):
                if res[k]:
                    res[k] = res[k].isoformat()

            if data:
                res['data'] = self.data
            return res

        def __repr__(self):
            return f'{self.category}-{self.id}-{self.user_id}'

    @app.route('/item/<category>', methods=['GET', 'POST'])
    @app.route('/item/<category>/<int:id_item>', methods=[
        'GET', 'PUT', 'PATCH', 'DELETE'
    ])
    @auth_required()
    def item(category, id_item=None):
        args = request.args
        method = request.method
        is_get = method == 'GET'
        kw = {'category': category, 'user_id': cu.id}
        if not is_get:
            kw['data'] = request.get_json()
        if 'name' in args:
            kw['name'] = args.get("name", type=str)

        by = {'category': category, 'user_id': cu.id}
        if id_item is not None:
            by['id'] = kw['id'] = id_item
        if is_admin():
            by.pop('user_id')
        if method == 'POST':  # Create.
            item = Item(**kw)
            db.session.add(item)
            db.session.flush()
            payload = item.payload()
        else:  # Read, Delete, Update/Modify, Update/Replace.
            query = Item.query.filter_by(**by)
            if id_item is None:  # GET
                query = query.order_by(Item.id)
                if 'page' in args and 'per_page' in args:
                    pag = db.paginate(
                        query,
                        page=args.get("page", type=int),
                        max_per_page=args.get("per_page", type=int),
                        count=True, error_out=False
                    )
                    items = [item.payload() for item in pag.items]
                    payload = {'page': pag.page, 'items': items,
                               'total': pag.total}
                else:
                    items = [item.payload() for item in query.all()]
                    payload = {'items': items, 'total': len(items)}
            else:
                item = query.first()
                if method == 'DELETE':
                    db.session.delete(item)
                elif method in ('PATCH', 'PUT'):
                    if method == 'PATCH':
                        kw['data'] = sh.combine_nested_dicts(
                            item.data, kw['data']
                        )
                    for k, v in kw.items():
                        setattr(item, k, v)
                    db.session.add(item)
                    db.session.flush()
                payload = item.payload(data=is_get)
        is_get or after_this_request(view_commit)
        return jsonify(payload)

    sitemap.add2csrf_protected(item=('view', 'item'))

    # Define models
    fsqla.FsModels.set_db_info(db)

    class Role(db.Model, fsqla.FsRoleMixin):
        pass

    class User(db.Model, fsqla.FsUserMixin):
        firstname = Column(String(255))
        lastname = Column(String(255))

        def get_security_payload(self):
            return {k: v for k, v in {
                'id': self.id,
                'email': self.email,
                'username': self.username,
                'firstname': self.firstname,
                'lastname': self.lastname
            }.items() if v is not None}

    # Setup Flask-Security
    class EditForm(Form):
        firstname = StringField('firstname', [Required()])
        lastname = StringField('lastname', [Required()])

    class ExtendedConfirmRegisterForm(ConfirmRegisterForm, EditForm):
        pass

    user_datastore = SQLAlchemyUserDatastore(db, User, Role)
    app.security = Security(
        app, user_datastore, confirm_register_form=ExtendedConfirmRegisterForm,
        register_blueprint=True
    )
    sitemap.add2csrf_protected(item=(
        'bp', app.config.get('SECURITY_BLUEPRINT_NAME')
    ))

    @app.route('/locales/<language>/<namespace>', methods=['GET'])
    def locales(language, namespace):
        from werkzeug.exceptions import NotFound
        from flask import send_from_directory

        for d in app.config['SCHEDULA_I18N_DIRNAME']:
            try:
                return send_from_directory(
                    d, f'{osp.join(language, "LC_MESSAGES", namespace)}.po',
                    as_attachment=True
                )
            except NotFound:
                pass
        raise NotFound()

    @app.route('/locales', methods=['GET'])
    def get_locales():
        return jsonify(current_app.config.get('BABEL_LANGUAGES'))

    @app.route(f'{app.config["SECURITY_URL_PREFIX"]}/edit', methods=['POST'])
    @auth_required()
    def edit():
        if request.is_json:
            data = MultiDict(request.get_json())
        else:
            data = request.form
        form = EditForm(data, meta=suppress_form_csrf())
        form.user = cu
        if form.validate_on_submit():
            after_this_request(view_commit)
            for k, v in form.data.items():
                setattr(cu, k, v)
            db.session.add(cu)
        return base_render_json(form)

    sitemap.add2csrf_protected(item=('view', 'edit'))

    class ContactForm(Form):
        name = StringField('name', [Required()])
        email = EmailField('email', [Required()])
        subject = StringField('subject', [Required()])
        message = StringField('message', [Required()])
        recaptcha = RecaptchaField('g-recaptcha-response')

    @app.route('/mail/contact', methods=['POST'])
    def contact():
        data = MultiDict(
            request.get_json()
        ) if request.is_json else request.form
        if cu.is_authenticated:
            if 'email' not in data:
                data['email'] = cu.email
            if 'name' not in data:
                data['name'] = f'{cu.firstname} {cu.lastname}'
        form = ContactForm(data, meta=suppress_form_csrf())
        if form.validate_on_submit():
            mail.send_rst(
                to=[form.data['email'],
                    current_app.config.get('MAIL_DEFAULT_SENDER')],
                rst='contact', reply_to=form.data['email'], user=cu, data=data,
                created=datetime.datetime.now().strftime("%d/%m/%Y-%H:%M:%S")
            )
            flash(
                str(lazy_gettext('Your message has been successfully sent!')),
                'success'
            )
        return base_render_json(form)

    def get_locale():
        from flask import request, session, current_app
        locale = session.get('locale')
        if not locale:
            session['locale'] = locale = request.accept_languages.best_match(
                current_app.config.get('BABEL_LANGUAGES')
            )
        return locale

    Babel(app, locale_selector=get_locale)
    mail = Mail(app)
    """
    account:
    accountID
    companyID(who
    this
    account
    belongs
    to)
    balanceAmount

    transactionType:
    transactionTypeID
    name(credit, debit, fee, etc.)
    transactionType(either - 1 or 1)

    transactionHistory:
    postDate,
    accountID,
    transactionTypeID,
    amount
    cat_toy = db.Table(
        "account_wallet_user",
        db.Base.metadata,
        Column('cat_id', ForeignKey('cats.id'), primary_key=True),
        Column('toy_id', ForeignKey('toys.id'), primary_key=True)
    )

    class AccountWallet(db.Model):
        __tablename__ = 'account_wallet'
        id = Column(Integer, primary_key=True)
        user_id = Column(Integer, ForeignKey('user.id'))
        created_at = Column(DateTime(), default=datetime.datetime.utcnow)
        updated_at = Column(DateTime(), onupdate=datetime.datetime.utcnow)

        def payload(self, data=False):
            res = {
                'id': self.id,
                'name': self.name,
                'created_at': self.created_at,
                'updated_at': self.updated_at
            }
            for k in ('created_at', 'updated_at'):
                if res[k]:
                    res[k] = res[k].isoformat()

            if data:
                res['data'] = self.data
            return res

        def __repr__(self):
            return f'{self.category}-{self.id}-{self.user_id}'
    """
    @app.route('/stripe/create-checkout-session', methods=['POST'])
    def create_payment():
        import stripe
        try:
            data = request.get_json() if request.is_json else dict(request.form)
            data = json_secrets.secrets(data, False)
            api_key = current_app.config.get('STRIPE_SECRET_KEY')

            if 'line_items' in data:
                it = data['line_items']
                if not isinstance(it, list):
                    it = [it]
                line_items = []
                for d in it:
                    if 'tiers' in d:
                        line_items.extend(compute_line_items(
                            d['quantity'], **d.pop('tiers')
                        ))
                    else:
                        line_items.append(d)
                lookup_keys = {}
                for i, d in enumerate(line_items):
                    lookup_key = d.pop('lookup_key', None)
                    if lookup_key:
                        sh.get_nested_dicts(
                            lookup_keys, lookup_key, default=list
                        ).append(i)
                if lookup_keys:
                    for price in stripe.Price.list(
                            api_key=api_key,
                            lookup_keys=list(lookup_keys.keys()),
                            expand=['data.product']
                    ).data:
                        for i in lookup_keys[price.lookup_key]:
                            line_items[i].update({'price': price.id})
                data['line_items'] = line_items

            session = stripe.checkout.Session.create(
                api_key=current_app.config.get('STRIPE_SECRET_KEY'),
                **sh.combine_nested_dicts(data, base={
                    'ui_mode': 'embedded',
                    'automatic_tax': {'enabled': True},
                    'redirect_on_completion': 'never',
                    'metadata': {
                        f'customer_{k}': getattr(cu, k)
                        for k in ('id', 'firstname', 'lastname')
                        if hasattr(cu, k)
                    }
                })
            )
        except Exception as e:
            return jsonify(error=str(e))

        return jsonify(
            clientSecret=session.client_secret, sessionId=session.id
        )

    @app.route('/stripe/session-status', methods=['GET'])
    def session_status():
        import stripe
        session = stripe.checkout.Session.retrieve(
            request.args.get('session_id'),
            api_key=current_app.config.get('STRIPE_SECRET_KEY')
        )
        status = session.status
        if status == "complete":
            msg = 'Payment succeeded!'
            category = 'success'
        elif status == "processing":
            msg = 'Your payment is processing.'
            category = 'success'
        elif status == "requires_payment_method":
            msg = 'Your payment was not successful, please try again.'
            category = 'success'
        else:
            msg = 'Something went wrong.'
            category = 'success'
        flash(str(lazy_gettext(msg)), category)
        return jsonify(
            status=status,
            customer_email=session.customer_details.email,
            userInfo=getattr(cu, "get_security_payload", lambda: {})()
        )

    @app.route('/stripe/webhook', methods=['POST'])
    def stripe_webhook():
        import stripe
        stripe.api_key = current_app.config.get('STRIPE_SECRET_KEY')
        payload = request.data
        sig_header = request.headers['STRIPE_SIGNATURE']

        try:
            event = stripe.Webhook.construct_event(
                payload, sig_header,
                current_app.config.get('STRIPE_WEBHOOK_SECRET_KEY'),
                tolerance=None
            )
        except ValueError as e:
            # Invalid payload
            raise e
        except stripe.error.SignatureVerificationError as e:
            # Invalid signature
            raise e
        sitemap.stripe_event_handler(event)

        return jsonify(success=True)

    stripe_webhook.csrf_exempt = True
    return app


def compute_line_items(quantity, tiers, type='graduated'):
    tiers = sorted(tiers, key=lambda x: x.get('last_unit', float('inf')))
    tiers[-1] = {k: v for k, v in tiers[-1].items() if k != 'last_unit'}
    line_items = []
    if type == 'volume':
        tier = next((
            tier for tier in tiers
            if quantity > tier.get('last_unit', float('inf'))
        ))
        if tier.get('flat_fee'):
            line_items.append({**tier['flat_fee'], 'quantity': 1})
        if tier.get('per_unit'):
            line_items.append({
                **tier['per_unit'],
                'quantity': quantity
            })
    else:
        prev_unit = 0
        for tier in tiers:
            last_unit = tier.get('last_unit', float('inf'))
            if tier.get('flat_fee'):
                line_items.append({**tier['flat_fee'], 'quantity': 1})
            if tier.get('per_unit'):
                line_items.append({
                    **tier['per_unit'],
                    'quantity': (min(last_unit, quantity) - prev_unit)
                })
            if quantity <= last_unit:
                break
            prev_unit = tier['last_unit']
    return line_items
